/*
 * XOR File Destroyer v.1 Beta 3
 *
 * AUTHOR:
 *  Robert Fahy <rfahy@ymail.com>
 *
 * DATE:
 *  May 2011 (initial release)
 *
 * ABOUT:
 *  - previous version was never released because it had entered the phase of "development hell",
 *   so I decided to begin anew and write a dumber version with the sole requirement that
 *   it "worked", whilst the process of improving robustness and security would follow
 *   only after a working version was out.
 *
 * TODO:
 *  - remove duplicate code
 *  - verify against using same file for input and output, in cases 2 and 3?
 *  - add wildcard operation
 *  - add verbosity (i.e. percent completed)
 *
 */

// include files
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// preprocessor definitions with obvious purposes
#define NAME_VERSION			"XOR-FD v.1 Beta 3"
#define GENERIC_FILENAME		"<xorcorrupt.elf>"
#define GENERIC_MESSAGE			"<Error text is unavailable.>"
#define BUFFER_SIZE				4096
#define STDIN_BUFFER_SIZE		4096

// ABOUT: according to the manpages this "will turn off_t into a 64-bit type".
//  translation: it allows us to work with larger files
#define _FILE_OFFSET_BITS		64

// evil global variables
char *_global_padding_ = NULL;
size_t _global_padding_length_ = 0;

// declarations without definitions (aka prototypes)
void print_usage(const char *program_filename, const char *message);
char * encrypt_static(char *data, size_t data_length);
char * encrypt_normal(char *data, const char *padding, size_t length);
int compute(const char *argument1, const char *argument2, const char *argument3);
int main(int argc, char **argv); // for completeness of this "index"

void print_usage(const char *program_filename, const char *message)
// ABOUT: prints the usage information and `message' to Standard Error stream (stderr)
{
#ifndef TEXT_BUFFER_SIZE
#define TEXT_BUFFER_SIZE	127
#endif

	char
		program_filename_[TEXT_BUFFER_SIZE + 1],
		message_[TEXT_BUFFER_SIZE + 1];

	if (program_filename != NULL &&
		message != NULL)
	{
		strncpy(program_filename_, program_filename, TEXT_BUFFER_SIZE);
		program_filename_[TEXT_BUFFER_SIZE] = '\0';
		strncpy(message_, message, TEXT_BUFFER_SIZE);
		message_[TEXT_BUFFER_SIZE] = '\0';
		fprintf(stderr, "\n%s usage:\n %s [file_to_be_encrypted] <[resulting_file] [file_used_as_key]>\n"
			"Also accepting standard input (stdin) data as key if `file_used_as_key' isn't provided.\n"
			"(Maximum stdin key length is %d bytes.)\n\nERROR: %s\n\n\n",
			NAME_VERSION, program_filename_, STDIN_BUFFER_SIZE, message_);
	}
} // ### void print_usage(const char *program_filename, const char *message)

char * encrypt_static(char *data, size_t length)
// ABOUT: uses XOR operation to pad contents of `data' with contents of `_global_padding_' for given `length'
// WARN: this version loops `_global_padding_' and it assumes that it doesn't change.
//  it also takes into account `_global_padding_length_'. this is a design flaw I'm aware of.
{
	static size_t i=0;

	if (data != NULL &&
		_global_padding_ != NULL)
		while (length--)
		{
			if (i >= _global_padding_length_)
			// reset `i' to keep within bounds
				i=0;

			*data++ ^= _global_padding_[i++];
		}
	else
		return NULL;

	return data;
} // ### char * encrypt_static(char *data, size_t data_length)

char * encrypt_normal(char *data, const char *padding, size_t length)
// ABOUT: uses XOR operation to pad contents of `data' with contents of `padding' for given `length'
// WARN: both arrays need to have same `length'
{
	if (data != NULL &&
		padding != NULL)
		while (length--)
			*data++ ^= *padding++;
	else
		return NULL;

	return data;
} // ### char * encrypt_normal(char *data, const char *padding, size_t length)

int compute(const char *argument1, const char *argument2, const char *argument3)
// ABOUT: verifies the arguments and takes the correct course of action
{
	FILE // file streams
		*input_file_ = NULL,
		*output_file_ = NULL,
		*key_file_ = NULL;

	char
		*rw_buffer_ = NULL,		// read/write file buffer
		*key_buffer_ = NULL;	// key file buffer

	size_t // to keep track of how many bytes we read or wrote
		rw_length_ = 0,
		key_length_ = 0,
		buffer_free_ = BUFFER_SIZE; // how much of the buffer is free (needed for case 3)

	unsigned long long int // to keep track of the total number of bytes we read or wrote
		total_bytes_read_ = 0,
		total_bytes_written_ = 0;

	if (argument1 == NULL)
	{
		print_usage(GENERIC_FILENAME, "Nice try.");
		return EXIT_FAILURE;
	}
	else
		if (argument2 == NULL)
		{
			if (argument3 == NULL)	// NOTE: case 1: input file overwritten, key from standard input
			{
				input_file_ = fopen(argument1, "r+b"); //! CHECK FOR ERROR
				_global_padding_ = (char *)calloc(STDIN_BUFFER_SIZE, sizeof (char)); //! CHECK FOR ERROR
				rw_buffer_ = (char *)calloc(BUFFER_SIZE, sizeof (char)); //! CHECK FOR ERROR

				// NOTE: we don't check this for errors because the program will jam waiting anyway,
				//  if no data is fed through standard input (stdin)
				_global_padding_length_ = fread(_global_padding_, sizeof (char), STDIN_BUFFER_SIZE, stdin);

				while ((rw_length_ = fread(rw_buffer_, sizeof (char), BUFFER_SIZE, input_file_))) //! CHECK FOR ERROR
				// did we read something? yes...
				{
					total_bytes_read_ += rw_length_;
					encrypt_static(rw_buffer_, rw_length_); //! CHECK FOR ERROR
					fseeko(input_file_, (off_t)-rw_length_, SEEK_CUR); //! CHECK FOR ERROR
					total_bytes_written_ += fwrite(rw_buffer_, sizeof (char), rw_length_, input_file_); //! CHECK FOR ERROR + ferror(rw_buffer_)
					fflush(input_file_);
				}

				fclose(input_file_); //! MOVE TO SPOT
				free(_global_padding_); //! MOVE TO SPOT
				_global_padding_ = NULL; //! MOVE TO SPOT
				free(rw_buffer_); //! MOVE TO SPOT
				rw_buffer_ = NULL; //! MOVE TO SPOT
			}
			else
			{
				print_usage(GENERIC_FILENAME, "Nice try.");
				return EXIT_FAILURE;
			}
		}
		else
			if (argument3 == NULL)	// NOTE: case 2: input file, output file, key from standard input
			{
				input_file_ = fopen(argument1, "rb"); //! CHECK FOR ERROR
				output_file_ = fopen(argument2, "wb"); //! CHECK FOR ERROR
				_global_padding_ = (char *)calloc(STDIN_BUFFER_SIZE, sizeof (char)); //! CHECK FOR ERROR
				rw_buffer_ = (char *)calloc(BUFFER_SIZE, sizeof (char)); //! CHECK FOR ERROR

				// NOTE: we don't check this for errors because the program will jam waiting anyway,
				//  if no data is fed through standard input (stdin)
				_global_padding_length_ = fread(_global_padding_, sizeof (char), STDIN_BUFFER_SIZE, stdin);

				while (!feof(input_file_))
				{
					total_bytes_read_ += (rw_length_ = fread(rw_buffer_, sizeof (char), BUFFER_SIZE, input_file_)); //! CHECK FOR ERROR
					encrypt_static(rw_buffer_, rw_length_); //! CHECK FOR ERROR
					total_bytes_written_ += fwrite(rw_buffer_, sizeof (char), rw_length_, output_file_); //! CHECK FOR ERROR
				}

				fclose(input_file_); //! MOVE TO SPOT
				fclose(output_file_); //! MOVE TO SPOT
				free(_global_padding_); //! MOVE TO SPOT
				_global_padding_ = NULL; //! MOVE TO SPOT
				free(rw_buffer_); //! MOVE TO SPOT
				rw_buffer_ = NULL; //! MOVE TO SPOT
			}
			else	// NOTE: case 3: input file, output file, key file
			{
				input_file_ = fopen(argument1, "rb"); //! CHECK FOR ERROR
				output_file_ = fopen(argument2, "wb"); //! CHECK FOR ERROR
				key_file_ = fopen(argument3, "rb"); //! CHECK FOR ERROR
				key_buffer_ = (char *)calloc(BUFFER_SIZE, sizeof (char)); //! CHECK FOR ERROR
				rw_buffer_ = (char *)calloc(BUFFER_SIZE, sizeof (char)); //! CHECK FOR ERROR

				while (!feof(input_file_))
				{
					key_length_ = 0;
					buffer_free_ = BUFFER_SIZE;
					total_bytes_read_ += (rw_length_ = fread(rw_buffer_, sizeof (char), BUFFER_SIZE, input_file_)); //! CHECK FOR ERROR

					while (key_length_ < rw_length_)
					// if we don't have enough padding data, read some more
					{
						key_length_ += fread(key_buffer_ + key_length_, sizeof (char), buffer_free_, key_file_); //! CHECK FOR ERROR
						buffer_free_ = BUFFER_SIZE - key_length_;

						if (feof(key_file_))	// if key_file_ was read entirely...
							rewind(key_file_);	// ... restart from the beginning
					}

					encrypt_normal(rw_buffer_, key_buffer_, rw_length_); //! CHECK FOR ERROR
					total_bytes_written_ += fwrite(rw_buffer_, sizeof (char), rw_length_, output_file_); //! CHECK FOR ERROR
				}

				fclose(input_file_); //! MOVE TO SPOT
				fclose(output_file_); //! MOVE TO SPOT
				fclose(key_file_); //! MOVE TO SPOT
				free(key_buffer_); //! MOVE TO SPOT
				key_buffer_ = NULL; //! MOVE TO SPOT
				free(rw_buffer_); //! MOVE TO SPOT
				rw_buffer_ = NULL; //! MOVE TO SPOT
			}

	printf("\nTotal bytes read: %llu\nTotal bytes written: %llu\n\n", total_bytes_read_, total_bytes_written_);

	return EXIT_SUCCESS;
} // ### int compute(const char *argument1, const char *argument2, const char *argument3)

// ### MAIN FUNCTION ###
int main(int argc, char **argv)
// ### MAIN FUNCTION ###
{
	if (argc > 4)
	{
		print_usage(argv[0], "Too many arguments.");
		return EXIT_FAILURE;
	}
	else
		if (argc < 2)
		{
			if (argc < 1)
			// NOTE: this shouldn't normally be possible except with a
			//  specially crafted exec() call from another program
			{
				print_usage(GENERIC_FILENAME, "Nice try.");
				return EXIT_FAILURE;
			}
			else
			{
				print_usage(argv[0], "The program needs at least one argument.");
				return EXIT_FAILURE;
			}
		}
		else
			switch (argc)
			{
				case 2: return compute(argv[1], NULL, NULL);		// case 1: input file overwritten, key from standard input
				case 3: return compute(argv[1], argv[2], NULL);		// case 2: input file, output file, key from standard input
				case 4: return compute(argv[1], argv[2], argv[3]);	// case 3: input file, output file, key file
				default:
					print_usage(argv[0], "Unknown. Email <rfahy@ymail.com> and include detail, please.");
					return EXIT_FAILURE;
			}
} // ### MAIN FUNCTION ###
